-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

NglPtDB = {};

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-- Required data saved for the Niggles Pet Teams addon
local requiredDB =
{
  layout       = {},
  petTeams     = {},
  pets         = {},
  settings     =
  {
    isVisible = false
  },
  filters      =
  {
    continents   = L.MASK_UNFILTERED,
    completeness = L.MASK_UNFILTERED,
    opponents    = L.MASK_UNFILTERED,
    lastEdited   = {},
    search       = L.MASK_UNFILTERED,
    text         = "",
  },
  tutorialSeen = 0,
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to create required settings. This function is recursive.
--
local function settingsCreate(required, current)
  -- Check each required setting exists
  for key, value in pairs(required) do
    -- Check if the required setting is a table
    if (type(value) == "table") then
      -- Create the required table, if needed
      if (type(current[key]) ~= "table") then
        current[key] = {};
      end

      -- Set any required settings for the current setting
      if (next(required[key])) then
        settingsCreate(value, current[key]);
      end
    elseif (type(value) ~= type(current[key])) then
      current[key] = value;
    end
  end

  return;
end

--
-- Function to destroy obsolete settings. This function is recursive.
--
local function settingsDestroy(required, current)
  -- Check the current setting isn't empty
  if (next(current)) then
    -- Check each of the current settings
    for key, value in pairs(current) do
      if (required[key] == nil) then
        current[key] = nil;
      elseif ((type(required[key]) == "table") and (next(required[key]))) then
        settingsDestroy(required[key], value);
      end
    end
  end

  return;
end

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to make sure all the settings are initialised
--
L.settingsInit = function()
  -- Destroy any obsolete data
  settingsDestroy(requiredDB, NglPtDB);

  -- Create any missing settings
  settingsCreate(requiredDB, NglPtDB);

  -- Make sure the tutorial seen flags are valid
  if (NglPtDB.settings.isVisible) then
    NglPtDB.tutorialSeen = bit.bor(NglPtDB.tutorialSeen,
      L.TUTORIALSEENFLAG_PETTEAMS);
  end
  if (next(NglPtDB.petTeams)) then
    NglPtDB.tutorialSeen = bit.bor(NglPtDB.tutorialSeen,
      L.TUTORIALSEENFLAG_PETTEAMS,
      L.TUTORIALSEENFLAG_NEWTEAM,
      L.TUTORIALSEENFLAG_TEAMEDIT);
  end

  -- Make sure the Last Edited settings are valid
  if (NglPtDB.filters.lastEdited[0] == nil) then
    NglPtDB.filters.lastEdited[0] = 1;
  end

  return;
end
